#############################################################################
##
#W  core_methods.gi Most common structural methods [quandles]
##
#H  @(#)$Id: core_methods.gi, v 2.1.3 2011/02/17 gap Exp $
##
#Y  Copyright (C)  2004,  G. P. Nagy (University of Szeged, Hungary),
#Y                        P. Vojtechovsky (University of Denver, USA)
##

#############################################################################
##  GENERATORS
##  -------------------------------------------------------------------------

#############################################################################
##
#A  GeneratorsOfRack( Q )
##
##  Returns a subset of <Q> gnerating <Q>. In most circumstances,
##  Elements( Q ) is returned.

InstallMethod( GeneratorsOfRack, "for rack",
    [ IsRack ],
    GeneratorsOfDomain
);



#############################################################################
##  COMPARING QUASIGROUPS WITH COMMON PARENT
##  -------------------------------------------------------------------------

#############################################################################
##
#A  \<( L, K )
##
##  Ordering between quasigroups with common parent quasigroup. Analogous to the
##  basic method for groups. We use the lexicographical ordering between the
##  small (greedy) generating sets.

InstallMethod(\<,
    "for racks by lexicographically ordered small generating sets",
    IsIdenticalObj,
    [IsRack, IsRack],
function(a,b)
    local ga,gb;
    ga:=GeneratorsSmallest(a);
    gb:=GeneratorsSmallest(b);
    if Length(ga)<Length(gb) then
        a:=Elements(a)[Size(a)];
        ga:=ShallowCopy(ga);
        while Length(ga)<Length(gb) do Add(ga,a); od;
    else
        b:=Elements(b)[Size(b)];
        gb:=ShallowCopy(gb);
        while Length(gb)<Length(ga) do Add(gb,b); od;
    fi;
    return ga<gb;
end);


#############################################################################
##  SECTIONS
##  -------------------------------------------------------------------------

#############################################################################
##
#A  LeftSection( Q )
##
##  Returns the left section of a quasigroup <Q>. Left section consists of
##  all left translations L_x: y --> x*y.

InstallMethod( LeftSection, "for rack",
    [ IsRack ],
function( Q )
    local P;
    P := PosInParent( Q );
    return List( CayleyTable( Q ), row -> MappingPermListList( P, row ) );
end );


#############################################################################
##
#O  LeftTranslation( Q, x )
##
##  Let <x> be an element of a quasigroup <Q>. Returns the left translation
##  by <x> in <Q>.

InstallMethod( LeftTranslation, "for quasigroup and quasigroup element",
    [ IsRack, IsRackElement ],
function( Q, x )
    if not x in Q then
        Error( "QUANDLES: <2> must be an element of the quasigroup <1>.");
    fi;
    return LeftSection( Q )[ Position( Q, x ) ];
end);


#############################################################################
##  MULTIPLICATION GROUPS
##  -------------------------------------------------------------------------

#############################################################################
##
#A  LeftMultiplicationGroup( Q )
##
##  Returns the subgroup of S_<Q> generated by all left translations of the
##  quasigroup <Q>.

InstallMethod( LeftMultiplicationGroup, "for rack",
    [IsRack ],
function( Q )
    return Group( LeftSection( Q ) );
end );




#############################################################################
##
#O  RelativeLeftMultiplicationGroup( L, S )
##
##  Returns the relative left multiplication group of a quasigroup <L>
##  with respect to its subquasigroup <S>.

InstallMethod( RelativeLeftMultiplicationGroup, "for two racks",
    [ IsRack, IsRack ],
function( L, S )
    local perms;
    if not IsSubrack( L, S ) then
        Error("QUANDLES: <2> must be a subrack of <1>.");
    fi;
    perms := List( S, x -> LeftTranslation( L, x ) );
    return Group( perms );
end);



#############################################################################
##
#A  InnerMappingGroup( L )
##
##  Returns the inner mapping group of a loop <L>, i.e., the subgroup
##  of MultplicationGroup( L ) consisting of maps fixing One( L ).

InstallMethod( DisplacementGroup, "for rack",
    [ IsRack ],
function( L )
	local perms;
    perms := List( L, x -> LeftQuotient(LeftTranslation( L, x ),LeftTranslation(L,Elements(L)[1])) );
    return Group( perms );
end);



#############################################################################
##  SUBQUASIGROUPS AND SUBLOOPS
##  -------------------------------------------------------------------------

InstallOtherMethod( Position, "for rack and rack element",
    [ IsRack, IsRackElement, IsInt],
function( Q, x, dummy )
    return Position( Elements( Q ), x );
end );

#############################################################################
##
#O  PosInParent( coll )
##
##  Let <coll> be a collection of elements of a quasigroup Q, and let P
##  be the parent of Q. Then PosInParent( coll )[ i ] is equal to
##  Pos ( Elements( P ), coll[ i ] ).
##  Note that we do not demand that all elements of <coll>
##  belong to the same quasigroup.

InstallMethod( PosInParent, "for collection of rack elements",
    [ IsCollection ],
function( coll )
    if not ForAll( coll, x -> IsRackElement( x ) ) then
        Error( "QUANDLES: <1> must be a list of rack elements." );
    fi;
    return List( coll, x -> x![ 1 ] );
end );

#############################################################################
##
#O  SubquasigroupNC( Q, pos_of_gens )
##
##  This auxiliary function assumes that:
##    a) Q is a quasigroup with Q = Parent( Q )
##    b) pos_of_gens is a nonempty subset of [ 1..Size( Q ) ] (not a sublist)
##    c) pos_of_gens determines a subquasigroup of Q
##  It then returns the corresponding subquasigroup of Q.

InstallMethod( SubrackNC, "for rack and set of integers",
    [ IsRack, IsSet ],
function( Q, pos_of_gens )
    local subqg, Qtype, elms;
    if IsEmpty( pos_of_gens ) then
        Error( "QUANDLES: <2> must be a nonempty subset.");
    fi;
    if not( Q = Parent( Q ) ) then
        Error( "QUANDLES: <1> must be its own parent." );
    fi;
    elms := Immutable( Elements( Q ){ pos_of_gens } );
    if IsQuandle( Q ) then
        Qtype := NewType( FamilyObj( elms ), IsQuandle and IsAttributeStoringRep );
    else
        Qtype := NewType( FamilyObj( elms ), IsRack and IsAttributeStoringRep );
    fi;
    subqg := Objectify( Qtype, rec( ) );
    SetSize( subqg, Length( pos_of_gens ) );
    SetAsSSortedList( subqg, elms );
    SetParent( subqg, Q );
    return subqg;
end );

#############################################################################
##
#O  Subquasigroup( Q, gens )
##
##  When <gens> is a nonempty list of elements of a quasigroup <Q>, or the
##  indices of them, it returns the subquasigroup of <Q> generated by <gens>.

InstallMethod( Subrack, "for quasigroup and list of elements",
    [ IsRack, IsList ],
function( Q, gens )
    local pos_gens, transl, pos_of_elms, relmultgr, subqg;
    # NG: we allow list of indices as well
    # PV: we allow gens to be empty, too. This will never happen when Subquasigroup is called from Subloop,
    # but it can happen when Q is a quasigroup. For instance, when Q has empty nucleus, it makes sense
    # to return [] rather than an error message when calling Nuc(Q).
    if IsEmpty( gens ) then
        return [];
    fi;
    if not( ForAll( gens, g -> g in Q ) or ForAll( gens, g-> g in PosInParent( Q ) ) ) then
        Error("QUANDLES: <2> must be a list of elements of quasigroup <1> or their indices.");
    fi;
    if not IsInt( gens[1] ) then
        pos_gens := PosInParent( gens );
    else
        pos_gens := ShallowCopy( gens );
        gens := Elements( Parent( Q ) ){ gens };
    fi;
    # P.V. ... must iterate this
    pos_of_elms := [];
    while pos_gens<>pos_of_elms do
        pos_of_elms := pos_gens;
        relmultgr := Subgroup( LeftMultiplicationGroup( Parent( Q ) ), LeftSection( Parent( Q ) ){ pos_gens } );
        pos_gens := Set( Orbits( relmultgr, pos_gens )[ 1 ] );
    od;
    subqg := SubrackNC( Parent( Q ), Set( pos_of_elms ) );

        SetGeneratorsOfMagma( subqg, gens );

    return subqg;
end );



#############################################################################
##
#O  IsSubquasigroup( Q, S )
##
##  Returns true if <S> is a subquasigorup of a quasigroup <Q>.

InstallMethod( IsSubrack, "for two quasigroups",
    [ IsRack, IsRack ],
function( Q, S )
    return Parent( Q ) = Parent( S ) and
           IsSubset( Elements( Q ), Elements( S ) );
end );

#############################################################################
##
#O  IsSubquandle( L, S )
##
##  Returns true if <S> is a subloop of a loop <L>.

InstallMethod( IsSubquandle, "for two loops",
    [ IsQuandle, IsQuandle ],
function( L, S )
    return IsSubrack( L, S );
end );

#############################################################################
##
#O  FixedPoints( A, f )
##
##  Returns the subset of A fixed by f

#InstallMethod ( FixedPoints, "for a group and an endomorphism",
#	[ IsGroup, IsMapping],
#function( A, f)
#	return Filtered(Elements(A), x-> x^f=x);
#end );


InstallMethod (AllHomomorphismsIncludingTrivial, "for two groups",
	[ IsGroup, IsGroup ],
function ( G, H )
	if Size(G)=1 then
		return [ MappingByFunction(G,H,x->Elements(H)[1]) ];
		else return AllHomomorphisms(G,H);
	fi;
end );

InstallMethod (AllEndomorphismsIncludingTrivial, "for a group",
	[ IsGroup ],
function ( G )
	if Size(G)=1 then
		return [ MappingByFunction(G,G,x->Elements(G)[1]) ];
		else return AllEndomorphisms(G);
	fi;
end );



#################################################################
##
#O  AffineMeshesOfGroups ( L );
##
##  Returns the list of all the affine meshes with orbits in the list L

homomorphism_table_search:=function ( gl, l, d,ii) 
	local fl,Tf,Tc,ht,i,j,hg,cl,ccag,
		add_to_list_of_tables,homomorphism_table_search_step,homomorphism_table_search_step_diagonal,constants_table_search_step;
	
add_to_list_of_tables :=function ()
	Add(Tf,Immutable(fl));
	Add(Tc,Immutable(cl));
end;

constants_table_search_step:=function(x,y)
	local c,i,j,k,v,ge;
	for c in Elements(gl[y]) do
		cl[x][y]:=c;
#		Print("Trying ",c," on position ",x,",",y,"\n");
		v:=true;
		for k in [1..y-1] do
			if v and Image(fl[y][k],c)<>Image(fl[k][k],cl[x][k]/cl[y][k]) then v:=false;fi;
		od;
		for j in [1..Minimum(y-1,x-1)] do
			if v and Image(fl[j][y],cl[x][j])<>Image(fl[y][y],c/cl[j][y]) then v:=false;fi;
		od;
		if (x<y) then
		for i in Union([1..x-1],[y]) do
			if v and Image(fl[x][y],cl[i][x])<>Image(fl[y][y],cl[i][y]/c) then v:=false;fi;
		od;fi;
		if v and (x=l)or((x=l-1)and(y=l)) then
			ge:=[];
			for i in [1..l] do
				Append(ge,AsList(Elements(Image(fl[i][y]))));
				Add(ge,cl[i][y]);
			od;
			if SubgroupNC(gl[y],ge)<>gl[y] then v:=false;
			fi;
		fi;
		if v then
#		Display (cl);
		j:=y;
		if (x+1=y) then k:=x+2; else k:=x+1; fi;
		if (k>l) then k:=1; j:=j+1; fi;
		if (j>l) then 
			add_to_list_of_tables ();
		else constants_table_search_step(k,j);
		fi;fi;
		cl[x][y]:=0;
	od;
end;

homomorphism_table_search_step:=function ( x, y) 
	local f,i,j,k,v;
	for f in ht[x][y] do
		fl[x][y]:=f;
		v:=true;
#		Print("Trying ",f," on position ",x,",",y,"\n");
		for k in [1..y] do
			if v and(CompositionMapping(fl[k][k],fl[x][k])<>CompositionMapping(fl[y][k],f)) then v:=false;fi;
		od;
		for j in [1..Minimum(x,y)] do
			if v and (CompositionMapping(fl[j][y],fl[x][j])<>CompositionMapping(fl[y][y],f)) then v:=false;fi;
		od;
		if (x<=y) then
			for i in Union([1..x],[y]) do
				if v and (CompositionMapping(f,fl[i][x])<>CompositionMapping(fl[y][y],fl[i][y])) then v:=false;fi;
			od;
		fi;
		if v then
#		Display (fl);
		j:=y;
		if (x+1=y) then k:=x+2; else k:=x+1; fi;
		if (k>l) then k:=1; j:=j+1; fi;
		if (j>l) then 
			constants_table_search_step (2,1);
		else homomorphism_table_search_step(k,j);
		fi;fi;
		fl[x][y]:=0;
	od;
end;

homomorphism_table_search_step_diagonal:=function ( x , dr ) 
	local f,dr2;
	for f in ht[x][x] do
		fl[x][x]:=f;
		if ii then
			dr2:=dr and Size(AsSet(List(Elements(Source(f)),z->Image(f,z))))=1;
		else 
			dr2:=dr and (f=ht[x][x][1]);
		fi;
		if (x<l) then homomorphism_table_search_step_diagonal(x+1,dr2);
		else if (l>1) then 
			if (not dr2) then homomorphism_table_search_step(2,1); fi;
			else if Size(Kernel(f))=1 then add_to_list_of_tables();
		fi;fi;fi;
	od;
end;



	fl:=List([1..l],x->List([1..l],x->0));
	cl:=List([1..l],x->List([1..l],x->0));
	Tf:=[];
	Tc:=[];
	ht:=List([1..l],x->List([1..l],x->0));
	for i in [1..l] do for j in [1..l] do
		if i<>j then ht[i][j]:=AllHomomorphismsIncludingTrivial(gl[i],gl[j]);
		else			
			if (ii) then 
				ht[i][i]:=[ MappingByFunction(gl[i],gl[i],x->x*x) ];
			else
				ccag:=ConjugacyClasses(AutomorphismGroup(gl[i]));
				ht[i][i]:=List(ccag,x->MappingByFunction(gl[i],gl[i],y->y/Image(Elements(x)[1],y)));
			fi;
			if (Exponent(gl[i])>d) then
				hg:=Filtered(ht[i][i], y-> Size(AsSet(List(Elements(gl[i]),z->Image(y,Image(y,z))^d)))=1);
#				Print("Seznam ",ht[i][i]," se zuzi na ",hg,"\n");
				ht[i][i]:=Immutable(hg);
			fi;
		fi;
	od;
		cl[i][i]:=Elements(gl[i])[1];
	od;
	homomorphism_table_search_step_diagonal(1,true);
	return(List([1..Length(Tf)],x->AffineMesh(gl,Tf[x],Tc[x])));
end;


InstallMethod( AffineMeshesOfGroups , "for a list of abelian groups ",
	[ IsList ],
function( L )
	return homomorphism_table_search(L,Length(L),Gcd(List(L,x->Exponent(x))),false);
end );
	
InstallMethod( InvolutoryAffineMeshesOfGroups , "for a list of abelian groups ",
	[ IsList ],
function( L )
	return homomorphism_table_search(L,Length(L),Gcd(List(L,x->Exponent(x))),true);
end );



InstallMethod( AffineMeshSupports , " for an integer ",
	[ IsInt , IsBool ],
function( n , al )
	local PS,p,lp,gl,k,l,aux,i;
        if al then
        PS:=Partitions(n);
        else
	PS:=Filtered(Filtered(Partitions(n),y->y[1]>2),x->Gcd(x)>1);
        fi;
	gl:=[];
	for p in PS do
		l:=Length(p);
		lp:=List([1..l],x->1);
		repeat
			aux:=[];
			for k in [1..l] do
				Add(aux,small_abelian_groups[p[k]][lp[k]]);
			od;
			Add(gl,Immutable(aux));
			k:=1;
			while (k<=l) and (lp[k]=Length(small_abelian_groups[p[k]])) do
				lp[k]:=1;
				k:=k+1;
			od;
			if k<=l then 
				lp[k]:=lp[k]+1;
				i:=k-1;
				while i>0 and p[k]=p[i] do
					if (lp[i]<lp[k]) then lp[i]:=lp[k];
					i:=i-1;
				fi;od;
			fi;
		until k>l;
	od;
	return (gl);
end );


InstallMethod( IsomorphismAffineMeshes , " for two affine meshes ",
	[ IsAffineMesh, IsAffineMesh ],
function ( M, N )
	local pi,psi,l,v,ag,find_next_aut,i,res,dl,K,cdl;
	
	find_next_aut:=function ( n )
		local i,v,f,d;
		for f in ag[n] do
			v:=true;
			psi[n]:=f;
#			Print("  a zkousim homomorfismus ",n," ",f);
			for i in [1..n] do
				v:=v and CompositionMapping(f,HomomorphismsMatrix(M)[i][n])=CompositionMapping(HomomorphismsMatrix(N)[i^pi][n^pi],psi[i]);
			od;
			if v then
				for  d in cdl[n^pi] do
#					Print( " d",n,"=",d," \n");
					dl[n]:=d;
					v:=true;
					for i in [1..n] do
						v:=v and ConstantsMatrix(M)[i][n]^f=ConstantsMatrix(N)[i^pi][n^pi]*(dl[i]^HomomorphismsMatrix(N)[i^pi][n^pi])/(d^HomomorphismsMatrix(N)[n^pi][n^pi]);
#						Print("   ? psi",n,"(c",i,n,")=c'",i^pi,n^pi,"+f'",i^pi,n^pi,"(",dl[i],")-f'",n^pi,n^pi,"(",d,") ..... ",v);
						v:=v and ConstantsMatrix(M)[n][i]^psi[i]=ConstantsMatrix(N)[n^pi][i^pi]*(d^HomomorphismsMatrix(N)[n^pi][i^pi])/(dl[i]^HomomorphismsMatrix(N)[i^pi][i^pi]);
#						Print("   ? psi",n,"(c",n,i,")=c'",n^pi,i^pi,"+f'",n^pi,i^pi,"(",dl[n],")-f'",i^pi,i^pi,"(",dl[i],") ..... ",v);
					od;
					if v then
						
#						Print(" ... vyslo\n");
						if n<l then find_next_aut(n+1);
						else res:=true;return true;
						fi;
#					else Print("... ne\n");
					fi;
				od;
			fi;
		od;
		return false;
	end;
	
	l:=Length(Orbits(M));
	if l<>Length(Orbits(N)) then return false;fi;
	if Reductivity(M)<>Reductivity(N) then return false;fi;
	res:=false;
	ag:=List(Orbits(M),AllAutomorphisms);
	cdl:=[1..l];
	for i in [1..l] do
		K:=Subgroup(Orbits(M)[i],Elements(Intersection(List(HomomorphismsMatrix(N)[i],x->Kernel(x)))));
		cdl[i]:=List(RightCosets(Orbits(M)[i],K),x->Representative(x));
#		cdl[i]:=Elements(Orbits(M)[i]);
	od;
	for pi in SymmetricGroup(l) do
		v:=true;
		for i in [1..l] do
			if v and Orbits(M)[i]<>Orbits(N)[i^pi] then v:=false;fi;
		od;
		if v then
		psi:=List([1..l],x->0);
		dl:=List([1..l],x->0);
#		Print(" Zkousim permutaci ",pi,"\n");
		find_next_aut(1);
#		Print(res);
		if res then return true;fi;
		fi;
	od;
	return res;
end );



InstallMethod(AffineMeshesIsomorphismClasses, "for a list of affine meshes",
	[IsList],
function( L )
	local il,i,j,v;
	il:=[];
	Print("next class\n");
	for i in [1..Length(L)] do
		v:=false;
		Print("...isomorphism search ",i/Length(L)+0.0,"                    %\r");
		for j in [1..Length(il)] do
			if not v then
				v:=IsomorphismAffineMeshes(il[j],L[i]);
			fi;
		od;
		if not v then Add(il,L[i]);
		fi;
	od;
	return il;
end );


InstallMethod(AffineMeshesOfSize, "for a number",
	[IsInt],
function(n)
	local gl,ml,il;
	gl:=AffineMeshSupports(n,false);
	ml:=List(gl,x->AffineMeshesOfGroups(x));
	il:=List(ml,x->AffineMeshesIsomorphismClasses(x));
	return il;
end );

InstallMethod(InvolutoryAffineMeshesOfSize, "for a number",
	[IsInt],
function(n)
	local gl,ml,il;
	gl:=AffineMeshSupports(n,false);
	ml:=List(gl,x->InvolutoryAffineMeshesOfGroups(x));
	il:=List(ml,x->AffineMeshesIsomorphismClasses(x));
	return il;
end );



InstallMethod(WriteMeshesToFile,"for a filename and a list of meshes",[IsString,IsList],
function(s,L)
   local M,fw,Q,T,i,j,l;
   fw:=OutputTextFile(s,false);
   for M in L do
      if Reductivity(M)=fail then AppendTo(fw,"Non");
      else AppendTo(fw,Reductivity(M));fi;
      l:=Size(M);
      AppendTo(fw,"-reductive medial quandle of size ",l,"\n   Orbits are ");
      for i in [1..Length(Orbits(M))] do
	 if i>1 then
	    if i<Length(Orbits(M)) then AppendTo(fw,", ");
	    else AppendTo(fw," and ");
	 fi;fi;
	 j:=1;
	 while Orbits(M)[i]<>small_abelian_groups[Size(Orbits(M)[i])][j]
	    do j:=j+1;od;
	 AppendTo(fw,small_abelian_group_names[Size(Orbits(M)[i])][j]);
      od;
      Q:=QuandleFromAffineMesh(M);
      AppendTo(fw,"\n   The left multiplication group is a group of order ",Size(LeftMultiplicationGroup(Q))," and exponent ",Exponent(LeftMultiplicationGroup(Q)));
      AppendTo(fw,"\n   The displacement group is a group of order ",Size(DisplacementGroup(Q))," and exponent ",Exponent(DisplacementGroup(Q)));
      T:=CayleyTable(Q);
      AppendTo(fw,"\n\n   |");
      for j in [1..l] do AppendTo(fw," ",Alfabeta[j]);od;
      AppendTo(fw,"\n---+-");
      for j in [1..l] do AppendTo(fw,"--");od;
      for i in [1..l] do 
	 AppendTo(fw,"\n ",Alfabeta[i]," |");
	 for j in [1..l] do AppendTo(fw," ",Alfabeta[T[i][j]]);od;
      od;
      AppendTo(fw,"\n\n\n");
   od;
   CloseStream(fw);
   return true;
end );


