#############################################################################
##
#W  quandles.gi  Representing, creating and displaying quandles [racks]
##
#H  @(#)$Id: quandles.gi, v 0.1.0 2013/03/04 gap Exp $
##
#Y  Copyright (C)  2013,  P. Jedlicka (CULS Prague)
##

#############################################################################
##  TESTING MULTIPLICATION TABLES
##  -------------------------------------------------------------------------

#############################################################################
##
#O  IsRackTable( ls )
##
##  Returns true if <ls> is an n by n left latin square with entries
##  1..n that is left distributive

InstallMethod( IsRackTable, "for matrix",
    [ IsMatrix ],
function( ls )
    local first_row,a,b,c;
    # checking rows
    first_row := Set( ls[ 1 ] );
    if not Length( first_row ) = Length( ls[ 1 ] ) then return false; fi;
    if ForAll( ls, row -> Set( row ) = first_row ) = false then return false; fi;
    # checking left distributivity
	for a in first_row do for b in first_row do for c in first_row do
		if ls[a][ls[b][c]] <> ls [ls[a][b]] [ls[a][c]] then return false; fi;
	od;od;od;
    return true;
end );

#############################################################################
##
#O  IsQuandleTable( ls )
##
##  Returns true if <ls> is a idempotent rack table

InstallMethod( IsQuandleTable, "for matrix",
    [ IsMatrix ],
function( ls )
	local a;
    if not IsRackTable( ls ) then return false; fi;
    for a in ls[1] do if ls[a][a]<>a then return false; fi; od;
	return true;
end );

#############################################################################
##  CREATING QUASIGROUPS AND LOOPS MANUALLY
##  -------------------------------------------------------------------------

#############################################################################
##
#A  CayleyTable( Q )
##
##  Returns the Cayley table of the rack <Q>

InstallMethod( CayleyTable, "for rack",
    [ IsRack ],
function( Q )
    local elms, parent_elms;
    elms := Elements( Q );
    parent_elms := Elements( Parent( Q ) );
    return List( elms, x-> List( elms, y -> Position( parent_elms, x * y ) ) );
end );

#############################################################################
##
#O  QuasigroupByCayleyTable( ct )
##
##  Returns rack with multiplication table <ct>.

InstallMethod( RackByCayleyTable, "for matrix",
    [ IsMatrix ],
function( ct )

    local F, Q, elms, n;
    if not IsRackTable( ct ) then
        Error( "QUANDLES: <1> must be a rack table." );
    fi;
    # constructing the family
    F := NewFamily( "RackByCayleyTableFam", IsRackElement );
    # installing data for the family
    n := Length ( ct );
    F!.size := n;
    elms := Immutable( List( [1..n], i -> Objectify(
        NewType( F, IsRackElement and IsRackElmRep), [ i ] ) ) );
    F!.set := elms;
    F!.cayleyTable := ct;
    F!.names := "q";
    # creating the quasigroup
    Q := Objectify( NewType( FamilyObj( elms ),
        IsRack and IsAttributeStoringRep ), rec() );
    # setting attributes for the quasigroup
    SetSize( Q, n );
    SetAsSSortedList( Q, elms );
    SetParent( Q, Q );
    SetCayleyTable( Q, ct );
    return Q;
end );

#############################################################################
##
#O  QuandleByCayleyTable( ct )
##
##  Returns quandle with multiplication table <ct>.

InstallMethod( QuandleByCayleyTable, "for matrix",
    [ IsMatrix ],
function( ct )
    local F, L, elms, n;
    if not IsQuandleTable( ct ) then
        Error( "QUANDLES: <1> must be a quandle table." );
    fi;
    # constructing the family
    F := NewFamily( "QuandleByCayleyTableFam", IsRackElement );
    # installing the data for the family
    n := Length ( ct );
    F!.size := n;
    elms := Immutable( List( [1..n], i -> Objectify(
        NewType( F, IsRackElement and IsRackElmRep), [ i ] ) ) );
    F!.set := elms;
    F!.cayleyTable := ct;
    F!.names := "q";
    # creating the loop
    L := Objectify( NewType( FamilyObj( elms ),
        IsQuandle and IsAttributeStoringRep ), rec() );
    # setting attributes for the loop
    SetSize( L, n );
    SetAsSSortedList( L, elms );
    SetParent( L, L );
    SetCayleyTable( L, ct );
    return L;
end );

#############################################################################
##
#O  SetQuasigroupElmName( Q, name )
##
##  Changes the name of elements of quasigroup or loop <Q> to <name>

InstallMethod( SetRackElmName, "for quasigroup and string",
    [ IsRack, IsString ],
function( Q, name )
    local F;
    F := FamilyObj( Elements( Q )[ 1 ] );
    F!.names := name;
end);




#############################################################################
##  DISPLAYING GROUPS AND LOOPS
##  -------------------------------------------------------------------------

InstallMethod( ViewObj, "for rack",
    [ IsRack ],
function( Q )
    Print( "<rack of order ", Size( Q ), ">" );
end );

## dangerous for large quasigroups
InstallMethod( PrintObj, "for rack",
    [ IsRack ],
function( Q )
    if HasCayleyTable( Q ) then
        Print( "<rack with multiplication table\n" );
        PrintArray( CayleyTable( Q ) );
    else
        Print( "<rack with elements\n" );
        Print( Elements( Q ) );
    fi;
    Print( ">\n" );
end );

InstallMethod( ViewObj, "for quandle",
    [ IsQuandle ],
function( L )
#    if HasIsAffineQuandle( L ) and IsAffineQuandle( L ) then
#        Print( "<affine quandle of order ", Size( L ), ">");
#    elif HasIsMedial( L ) and IsMedial( L ) then
#        Print( "<medial quandle of order ", Size( L ), ">");
#    elif HasIsAbelianQuandle( L ) and IsAbelianQuandle( L ) then
#        Print( "<Abelian quandle of order ", Size( L ), ">");
#    else
        # MORE ??
        Print( "<quandle of order ", Size( L ), ">" );
#    fi;
end );

## dangerous for large loops
InstallMethod( PrintObj, "for quandle",
    [ IsQuandle ],
function( L )
    if HasCayleyTable( L ) then
        Print( "<quandle with multiplication table\n" );
        PrintArray( CayleyTable( L ) );
    else
        Print( "<quandle with elements\n" );
        Print( Elements( L ) );
    fi;
    Print( ">\n" );
end );

InstallMethod( ViewObj, "for abelian mesh",
    [ IsAffineMesh ],
function( Q )
   if HasReductivity (Q) then
      if Reductivity(Q)=fail then
	 Print( "< NR affine mesh of size ", Size( Q ), " with ", Size( Orbits(Q) ) , " orbits>" );
      else
	 Print( "<",Reductivity(Q),"R affine mesh of size ", Size( Q ), " with ", Size( Orbits(Q) ) , " orbits>" );
      fi;
   else
      Print( "<affine mesh of size ", Size( Q ), " with ", Size( Orbits(Q) ) , " orbits>" );
   fi;
end );

############################################################
##
#O  Galkin Representation
##


InstallMethod(GalkinRepresentation,"for group its subgroup and an automorphism",
	[IsGroup,IsGroup,IsMapping],
function (G,H,f)
	local C,T,n,a,b,x,y,z,u;
	if not IsSubgroup(G,H) then
		Error("QUANDLES: H has to be a subgroup of G");
	fi;
	for x in H do
	if not x^f=x then
		Error("QUANDLES: H has to be fixed by f");
	fi;od;
	C:=RightCosets(G,H);
	n:=Length(C);
	T:=List([1..n],x->[1..n]);
	for a in [1..n] do for b in [1..n] do
		x:=Representative(C[a])^(-1);
		y:=Representative(C[b])^(-1);
		z:=x*((x^(-1)*y)^f);
		u:=RightCoset(H,z^(-1));
		T[a][b]:=Position(C,u);
	od;od;
	return QuandleByCayleyTable(T);
end);





InstallMethod(MapNilpotency,"for en endpmorphism",[IsMapping],
function(f)
   local sF,i,g,zm;
   i:=0;
   sF:=Set([]);
   zm:=MappingByFunction(Source(f),Range(f),x->Elements(Range(f))[1]);
   g:=IdentityMapping(Source(f));
   while 0<1 do
      i:=i+1;
      g:=CompositionMapping(g,f);
      if g=zm then return i; fi;
      if g in sF then return fail;fi;
      AddSet(sF,g);
   od;
end);






#########################################################
##  AFFINE MESHES
##
#
#InstallMethod(Orbits," for an affine mesh ", 
#	[IsAffineMesh],
#function( M ) 
#	return Elements( M );
#end );

InstallMethod(AffineMesh, " for a list ",
	[ IsList , IsList , IsList ],
function( gl , fl , cl)
	local g,s,M,ci,i,j,l,k;
	s:=0;
	l:=Length(gl);
	if Length(cl)<>l then
		Error("QUANDLES: Sizes do not match");
	fi;
	if Length(fl)<>l then
		Error("QUANDLES: Sizes do not match");
	fi;
	for g in gl do if (not IsGroup(g)) or (not IsAbelian(g)) then
		Error( "QUANDLES: We need a list of abelian groups");
		else s:=s+Size(g);
	fi;od;
	for ci in cl do
		if Length(ci)<>l then
			Error("QUANDLES: Sizes do not match");
		fi;
		for i in [1..l] do if not ci[i] in Elements(gl[i]) then
			Error("QUANDLES: Constants c_ij must lie in the group A_j");
		fi;od;
	od;
	for i in [1..l] do if cl[i][i]<>Elements(gl[i])[1] then
		Error("QUANDLES: Constants c_ii must be 0");
	fi;od;
	for j in [1..l] do
		if Length(fl[j])<>l then
			Error("QUANDLES: Sizes do not match");
		fi;
		for i in [1..l] do if (not IsMapping(fl[j][i])) or (Range(fl[j][i])<>gl[i]) or (Source(fl[j][i])<>gl[j]) then
			Error("QUANDLES: f_ij must be a mapping from A_i to A_j");
		fi;od;
	od;
	for i in [1..l] do if Size(Filtered(Elements(gl[i]), x-> Image(fl[i][i],x)=x))>1 then
		Error("QUANDLES: Mappings f_ii must be fixed point free");
	fi;od;
	for i in [1..l] do for j in [1..l] do for k in [1..l] do
		if (CompositionMapping(fl[i][j],fl[k][i])<>CompositionMapping(fl[j][j],fl[k][j])) then
			Error("QUANDLES: Mappings f must satisfy a commuting relation");
		fi;
		if Image(fl[j][k],cl[i][j])<>Image(fl[k][k],cl[i][k]/cl[j][k]) then
			Error("QUANDLES: Mappings f and constants c are not suitable");
		fi;
	od;od;od;
	M := Objectify( NewType( FamilyObj( gl ),
        IsAffineMesh and IsAttributeStoringRep ), rec() );
    SetSize(M,s);
    SetOrbits(M,gl);
    SetAsSSortedList(M,gl);
    SetConstantsMatrix(M,cl);
    SetHomomorphismsMatrix(M,fl);
    SetParent(M,M);
    return M;
end );

InstallMethod(AffineMeshNoCheck, " for a list ",
	[ IsList , IsList , IsList , IsInt],
function( gl , fl , cl, s)
	local M;
	M := Objectify( NewType( FamilyObj( gl ),
        IsAffineMesh and IsAttributeStoringRep ), rec() );
    SetSize(M,s);
    SetOrbits(M,gl);
    SetAsSSortedList(M,gl);
    SetConstantsMatrix(M,cl);
    SetHomomorphismsMatrix(M,fl);
    SetParent(M,M);
    return M;
end );



InstallMethod(QuandleFromAffineMesh," for an affine mesh ",
	[IsAffineMesh],
function ( M ) 
	local n,T,S,i,j,l,x,y;
	l:=Length(Orbits(M));
	n:=Size(M);
	S:=[1..l];
	S[1]:=0;
	for i in [2..l] do
		S[i]:=S[i-1]+Size(Orbits(M)[i-1]);
	od;
	T:=List([1..n],x->List([1..n],x->0));
	for i in [1..l] do for j in [1..l] do for x in [1..Size(Orbits(M)[i])] do for y in [1..Size(Orbits(M)[j])] do
		T[S[i]+x][S[j]+y]:=S[j]+Position(Elements(Orbits(M)[j]),Image(HomomorphismsMatrix(M)[i][j],Elements(Orbits(M)[i])[x]) * 
			(Elements(Orbits(M)[j])[y]) * Image(HomomorphismsMatrix(M)[j][j],((Elements(Orbits(M)[j])[y])^(-1))) * ConstantsMatrix(M)[i][j]);
	od;od;od;od;
	return (QuandleByCayleyTable(T));
end );

InstallMethod(Reductivity," for an affine mesh ",[IsAffineMesh],
function(M)
   local i,l,n,k;
   l:=Length(Orbits(M));
   if l=Size(M) then return 1;fi;
   n:=0;
   for i in [1..l] do
      k:=MapNilpotency(HomomorphismsMatrix(M)[i][i]);
      if k=fail then return fail;fi;
      if k+1>n then n:=k+1;fi;
   od;
   return n;
end );

InstallMethod(TwoReductiveConstantMatrix," for a number and a list od groups ",[IsInt,IsList],
function(n,L)
   local i,j,l,M,k;
   l:=Length(L);
   M:=List([1..l],x->[1..l]);
   for i in [1..l] do
      for j in [1..l] do
	 if i=j then M[i][j]:=Elements(L[j])[1];
	 else
	    k:=n mod Size(L[j]);
	    M[i][j]:=Elements(L[j])[k+1];
	    n:=(n-k)/Size(L[j]);
	 fi;
      od;
   od;
   return M;
end );

InstallMethod(TwoReductiveConstantMatrixInverse," for a matrix and a list od groups ",[IsList,IsList],
function(M,L)
   local i,j,l,n,k;
   l:=Length(L);
   n:=0;
   for i in Reversed([1..l]) do
      for j in Reversed([1..l]) do
	 if i<>j then 
	    n:=n*Size(L[j]);
	    n:=n+Position(Elements(L[j]),M[i][j])-1;
	 fi;
      od;
   od;
   return n;
end );

InstallMethod(PermutationsOfOrderedGroupList," for a list of groups ",[IsList],
function(L)
   local PL,i;
   PL:=[()];
   for i in [2..Length(L)] do
      if L[i]=L[i-1] then Add(PL,(i-1,i));fi;
   od;
   return Group(PL);
end );

InstallMethod(LessAffineMeshRow," for a list of groups and two matrix rows ",[IsList,IsList,IsList],
function(L,r1,r2)
	local i,k;
	for i in [1..Length(L)] do
		k:=Position(Elements(L[i]),r1[i])-Position(Elements(L[i]),r2[i]);
		if k<>0 then return k;fi;
	od;
	return 0;
end );

InstallMethod(IsSortedConstantMatrix," for a matrix and a row index",[IsList,IsInt,IsList],
function(M,n,L)
	local i,k;
	k:=Length(M)-1;
	for i in [n..k] do
		if LessAffineMeshRow(L,M[i],M[i+1])<0 then return false;fi;
	od;
	return true;
end );


InstallMethod(AllTwoReductive," for a list of groups ",[IsList],
function(L)
   local n,i,j,l,ot,CL,JeTam,PG,p,MT,NMT,aNMT,cCL,conn,PSI,psi,S1,aux;
   l:=Length(L);
   CL:=[];
   PG:=PermutationsOfOrderedGroupList(L);
   if Size(PG)=1 then S1:=0;
   else
      i:=1;
      JeTam:=true;
      S1:=0;
      while JeTam and i+1<l do
	if Size(L[i])=1 then 
	   JeTam:=false;
	   S1:=i;
	else if L[i]=L[i+1] then
	   JeTam:=false;
	fi;fi;
	i:=i+1;
     od;
   fi;
   PSI:=DirectProduct(List(L,AutomorphismGroup));
   ot:=Product(L,Size)^(l-1);
   for n in [1..ot] do
      JeTam:=false;
      j:=1;
      while (not JeTam)and(j<=Length(CL)) do
	 JeTam:=n in CL[j];
	 j:=j+1;
      od;
      if not JeTam then
	 MT:=TwoReductiveConstantMatrix(n,L);
	 conn:=true;
	 j:=1;
	 while conn and (j<=l) do
	    conn:=(SubgroupNC(L[j],List([1..l],x->MT[x][j]))=L[j]);
	    j:=j+1;
	 od;
	 if conn then
	    cCL:=Set([]);
	    if S1=0 then
	       for p in PG do
		  NMT:=List([1..l],x->List([1..l],y->MT[x^p][y^p]));
		  if not (TwoReductiveConstantMatrixInverse(NMT,L) in cCL) then
		     for psi in PSI do
			aNMT:=List([1..l],x->List([1..l],y->Image(psi[y],NMT[x][y])));
			AddSet(cCL,TwoReductiveConstantMatrixInverse(aNMT,L));
		     od;
		  fi;
	       od;
	    else
	       for psi in PSI do
		  aNMT:=List([1..l],x->List([1..l],y->Image(psi[y],MT[x][y])));
		  for i in [S1..l-1] do for j in [i+1..l] do
		     if LessAffineMeshRow(L,aNMT[i],aNMT[j])<0 then
			aux:=aNMT[i];
			aNMT[i]:=aNMT[j];
			aNMT[j]:=aux;
		  fi;od;od;
		  AddSet(cCL,TwoReductiveConstantMatrixInverse(aNMT,L));
	       od;
	       if Length(cCL)>0 then
		  for n in [1..ot] do
		     JeTam:=false;
		     j:=1;
		     while (not JeTam)and(j<=Length(CL)) do
			JeTam:=cCL[1] in CL[j];
			j:=j+1;
		     od;
		  od;
		  if JeTam then cCL:=[];fi;
	       fi;
	    fi;
	    if Length(cCL)>0 then Add(CL,AsSortedList(cCL));fi;
	 fi;
      fi;
   od;
   return CL;
end );

InstallMethod(NumberAffineMeshesZ2Z1," for the number of Z2 and Z1 orbits ",[IsInt,IsInt],
function(n,m)

local g,h,rg,rh,p,s,cycp,cycs,sum,prod,c;

g:=SymmetricGroup(n);
h:=SymmetricGroup(m);

rg:=List(ConjugacyClasses(g),x->Representative(x));
rh:=List(ConjugacyClasses(h),x->Representative(x));

sum:=0;
for p in rg do for s in rh do
cycp := List(Cycles(p, [1..n]), x->Size(x));
cycs := List(Cycles(s, [1..m]), x->Size(x));
prod:=1;
for c in cycp do
prod := prod * (2^( Sum(List(cycp,x->Gcd(c,x))) + Sum(List(cycs,x->Gcd(c,x))) -1) -1);
od;
sum := sum + Size(ConjugacyClass(g,p)) * Size(ConjugacyClass(h,s)) * prod;
od;od;

return sum/(Factorial(n)*Factorial(m));

end );



small_abelian_groups:=[
[AbelianGroup([1])],
[AbelianGroup([2])],
[AbelianGroup([3])],
[AbelianGroup([4]),AbelianGroup([2,2])],
[AbelianGroup([5])],
[AbelianGroup([2,3])],
[AbelianGroup([7])],
[AbelianGroup([8]),AbelianGroup([4,2]),AbelianGroup([2,2,2])],
[AbelianGroup([9]),AbelianGroup([3,3])],
[AbelianGroup([2,5])],
[AbelianGroup([11])],
[AbelianGroup([4,3]),AbelianGroup([3,2,2])],
[AbelianGroup([13])],
[AbelianGroup([7,2])],
[AbelianGroup([5,3])],
[AbelianGroup([16]),AbelianGroup([8,2]),AbelianGroup([4,4]),AbelianGroup([4,2,2]),AbelianGroup([2,2,2,2])],
[AbelianGroup([17])],
[AbelianGroup([9,2]),AbelianGroup([3,3,2])],
[AbelianGroup([19])],
[AbelianGroup([5,4]),AbelianGroup([5,2,2])],
[AbelianGroup([7,3])],
[AbelianGroup([11,2])],
[AbelianGroup([23])]];

small_abelian_group_names:=[
["Z_1"],
["Z_2"],
["Z_3"],
["Z_4","Z_2^2"],
["Z_5"],
["Z_6"],
["Z_7"],
["Z_8","Z_4xZ_2","Z_2^3"],
["Z_9","Z_3^2"],
["Z_10"],
["Z_11"],
["Z_12","Z_3xZ_2^2"],
["Z_13"],
["Z_14"],
["Z_15"],
["Z_16","Z_8xZ_2","Z_4^2","Z_4xZ_2^2","Z_2^4"],
["Z_17"],
["Z_18","Z_3^2xZ_2"],
["Z_19"],
["Z_20","Z_5xZ^2^2"],
["Z_21"],
["Z_22"],
["Z_23"]];

Alfabeta:=["a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z"];
